/* 
    qasn1 - Asn1BerParser - parser class for qasn1 library
    Copyright(C) 2010 - Andrius da Costa Ribas

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Asn1BerParser.h"
#include <QtCore/QDebug>

class Asn1BerParserPrivate
{
public:
    Asn1BerParserPrivate(QIODevice* inputStream, Asn1BerParser* parentParser);
    Asn1BerParserPrivate(QIODevice* inputStream, int endOffset);
    Asn1BerParserPrivate(QByteArray inputStream);
    
    QIODevice* m_inputDevice;
    QByteArray m_parsingStream;
    int m_maxOffset;
    
    // if the input is a QByteArray then m_inputDevice should be deleted on destructor
    bool m_useBuffer;
    bool m_enableProgressDialog;
    QProgressDialog* m_progressDialog;
};

Asn1BerParserPrivate::Asn1BerParserPrivate(QIODevice* inputStream, Asn1BerParser* parentParser)
{
    m_inputDevice = inputStream;
    if ( parentParser == 0 )
    {
        m_maxOffset = m_inputDevice->size();
    }
    else
    {
        m_maxOffset = parentParser->maxOffset();
    }
    
    m_useBuffer = false;
    m_enableProgressDialog = false;
    m_progressDialog = 0;
}

Asn1BerParserPrivate::Asn1BerParserPrivate(QIODevice* inputStream, int endOffset)
{
    m_inputDevice = inputStream;
    m_maxOffset = endOffset;
    
    m_useBuffer = false;
    m_enableProgressDialog = false;
    m_progressDialog = 0;
}

Asn1BerParserPrivate::Asn1BerParserPrivate(QByteArray inputStream)
{
    QBuffer* buffer = new QBuffer;
    buffer->buffer() = inputStream;
    buffer->open(QIODevice::ReadOnly);
    m_inputDevice = buffer;
    m_maxOffset = m_inputDevice->size();
    
    m_useBuffer = true;
    m_enableProgressDialog = false;
    m_progressDialog = 0;
}

Asn1BerParser::Asn1BerParser(QIODevice* inputStream, Asn1BerParser* parentParser) : d( new Asn1BerParserPrivate( inputStream, parentParser ) )
{
    if ( parentParser != 0 )
    {
        d->m_progressDialog = parentParser->progressDialog();
    }
}

Asn1BerParser::Asn1BerParser(QIODevice* inputStream, int endOffset) : d( new Asn1BerParserPrivate( inputStream, endOffset ) )
{
}

Asn1BerParser::Asn1BerParser(QByteArray inputStream) : d( new Asn1BerParserPrivate( inputStream ) )
{
}

QProgressDialog* Asn1BerParser::progressDialog()
{
    return d->m_progressDialog;
}

int Asn1BerParser::maxOffset()
{
    return d->m_maxOffset;
}

void Asn1BerParser::enableProgressDialog(bool enabled)
{
    d->m_enableProgressDialog = enabled;
}

bool Asn1BerParser::progressDialogEnabled()
{
    return d->m_enableProgressDialog;
}


Asn1BerNode* Asn1BerParser::parse(Asn1BerNode* parentNode)
{
    char currentOctet;
    char tagClass;
    unsigned int tagNumber;
    unsigned int length;
    unsigned int lenOfLen;
    Asn1BerNode* currentNode;
    Asn1BerNode* firstSibling;
    Asn1BerNode* previousNode;
    bool constructed;

    firstSibling = 0;
    previousNode = 0;
    
    Q_ASSERT(d->m_inputDevice->isReadable());
    
    if ( progressDialogEnabled() )
    {
        if ( d->m_progressDialog == 0 )
        {
            d->m_progressDialog = new QProgressDialog(0, Qt::Dialog);
            d->m_progressDialog->setLabelText(QString("Parsing Asn.1 BER file"));
            d->m_progressDialog->setCancelButton(0);
            d->m_progressDialog->setMinimum(0);
            d->m_progressDialog->setMaximum(maxOffset());
            d->m_progressDialog->setWindowModality(Qt::WindowModal);
        }
    }
    
    //qDebug() << "one level down";

    while ( d->m_inputDevice->bytesAvailable() >= 0 && d->m_inputDevice->pos() <= d->m_maxOffset )
    {
        if ( d->m_progressDialog != 0 )
        {
            // Only top level sets enabled flag, but it may share the progressDialog with children parsers
            d->m_progressDialog->setValue(d->m_progressDialog->maximum() - d->m_inputDevice->bytesAvailable());
        }
        if (d->m_inputDevice->bytesAvailable() == 0)
        {
            //qDebug() << "EOF";
            if ( parentNode != 0 && !(parentNode->hasDefinedLength()) )
            {
                //qDebug() << "Unexpected EOF, undefined length parent";
                currentNode = new Asn1BerNode;
                currentNode->setError("No end-of-contents octet inside unknown length node, found EOF instead");
                if ( previousNode != 0 )
                {
                    previousNode->setNext(currentNode);
                    currentNode->setPrev(previousNode);
                }
            }
            return firstSibling;
        }

        currentNode = new Asn1BerNode();
        if ( firstSibling == 0 )
        {
            //qDebug() << "first node on loop";
            firstSibling = currentNode;
        }
        else
        {
            //qDebug() << "new node on loop";
            currentNode->setPrev(previousNode);
            previousNode->setNext(currentNode);
        }
    
        ////////////////////// Parsing TAG
        
        if ( !(d->m_inputDevice->getChar(&currentOctet)) )
        {
            //qDebug() << "Premature EOF";
            currentNode->setError("Unexpected EOF parsing TAG");
            return firstSibling;
        }
    
        // the two first bits correspond to the class of tag:
        tagClass = (unsigned char) currentOctet & TWO_FIRST_BITS;
        
        currentNode->setClass(tagClass);
        
        //qDebug() << "Tag Class: " << ((unsigned int) tagClass);
    
        // the third bit indicates whether it is primitive or complex
        constructed = ( ( currentOctet & THIRD_BIT ) == CODING_COMPLEX );
        if ( constructed )
        {
            //qDebug() << "Constructed";
            currentNode->setCodingComplex();
        }
    
        // if the remaining bits are not 11111, then encoding of tag is in a single octet
        tagNumber = currentOctet & LAST_FIVE_BITS;
        
        // case it is not the tag number is in the seven last bits of
        // following octets until an octet has a MSB of zero
        if ( tagNumber == BIG_TAG_IND )
        {
            //qDebug() << "Big Tag";
            tagNumber = 0;
            do
            {
                if ( !(d->m_inputDevice->getChar(&currentOctet)) )
                {
                    //qDebug() << "premature EOF";
                    currentNode->setError("Unexpected EOF parsing TAG");
                    return firstSibling;
                }
                tagNumber *= 0x80; // shifting 7 bits (do not trust much in [<< 7])
                tagNumber += ( currentOctet & LAST_SEVEN_BITS );
            }while(currentOctet & FIRST_BIT);
        }
        currentNode->setTag(tagNumber);
        //qDebug() << "Tag: " << tagNumber;
    
        ////////////////////// Parsing LENGTH
        
        if ( !(d->m_inputDevice->getChar(&currentOctet)) )
        {
            //qDebug() << "premature EOF";
            currentNode->setError("Unexpected EOF parsing LENGTH");
            return firstSibling;
        }
    
        // 0x80 denotes unknown length
        // node's default length state is unknown
        if ( (unsigned char) currentOctet != UNKNOWN_LENGTH )
        {
            if ( currentOctet & BIG_LEN_IND )
            {
                //qDebug() << "Big Length";
                lenOfLen = (unsigned char) currentOctet & ~BIG_LEN_IND;
                length = 0;
                for ( register unsigned int i=0; i<lenOfLen; i++ )
                {
                    if ( !(d->m_inputDevice->getChar(&currentOctet)) )
                    {
                        currentNode->setError("Unexpected EOF parsing LENGTH");
                        return firstSibling;
                    }
                    length = length * 0x100 + (unsigned char) currentOctet;
                }
            }
            else
            {
                length = currentOctet;
            }
            //currentNode->setLen(length);
            //qDebug() << "Length: " << length;
            currentNode->setDefinedLength(true);
        }
        else
        {
            //qDebug() << "Undefined Length";
            length = -1; // just to not leave uninitialized;
        }
    
        // if tag and length are zero, and parent has undefined length then parent data stops here
        if ( tagNumber == 0 && length == 0 && tagClass == CLASS_UNIVERSAL)
        {
            if ( parentNode != 0 )
            {
                if ( !(parentNode->hasDefinedLength()) )
                {
                    //qDebug() << "end-of-contents, one level up";
                    // previousNode is linked to currentNode, we must remove currentNode
                    if ( currentNode == firstSibling )
                    {
                        // dangling pointers!!
                        firstSibling = 0;
                    }
                    delete currentNode;
                    return firstSibling;
                }
            }
        }
    
        ////////////////////// Parsing VALUE
    
        // no value if node has subnodes
        if ( currentNode->isCodingPrimitive() )
        {
            //qDebug() << "Primitive";
            if ( d->m_inputDevice->bytesAvailable() < length )
            {
                //qDebug() << "premature EOF";
                currentNode->setError("Unexpected EOF parsing VALUE");
                return firstSibling;
            }

            currentNode->setVal(d->m_inputDevice->read(length));
            
            previousNode = currentNode;
            //qDebug() << "Value set: " << (currentNode->val());
            
        }
        else if ( currentNode->hasDefinedLength() )
        {
            //qDebug() << "Complex, defined length";
            Asn1BerParser* childParser;
            Asn1BerNode* childNode;
            
            childParser = new Asn1BerParser(d->m_inputDevice, length + d->m_inputDevice->pos() - 1);
            childNode = childParser->parse(currentNode);
            delete childParser;
            
            if ( childNode != 0 )
            {
                currentNode->setChild(childNode);
                childNode->setParent(currentNode);
            }
            
            previousNode = currentNode;
        }
        else
        {
            //qDebug() << "Complex, undefined length";
            Asn1BerParser* childParser;
            Asn1BerNode* childNode;        
            
            childParser = new Asn1BerParser(d->m_inputDevice, this);
            childNode = childParser->parse(currentNode);
            delete childParser;
            
            if ( childNode != 0 )
            {
                currentNode->setChild(childNode);
                childNode->setParent(currentNode);
            }
            
            previousNode = currentNode;
        }
        
    }
    return firstSibling;
}

Asn1BerParser::~Asn1BerParser()
{
    // only top level can enable and destroy progress dialog
    if ( d->m_enableProgressDialog )
    {
        delete d->m_progressDialog;
    }
    delete d;
}
