/* 
    qasn1 - parsertest - unit tests for parser class of qasn1 library
    Copyright(C) 2010 - Andrius da Costa Ribas

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QtCore/QDebug>
#include <QtCore/QObject>
#include <QtTest/QtTest>

#include "Asn1BerParser.h"

#include "parsertest.h"

void Asn1BerParserTest::initTestCase()
{
    qDebug() << "startup";
}

void Asn1BerParserTest::testPrimitive()
{
    QByteArray stream;
    
    stream.append((char) 0x4c); // Application [12] primitive
    stream.append((char) 0x02); // length = 2
    stream.append('a'); // 1st byte of value
    stream.append('b'); // 2nd byte of value
    
    Asn1BerParser* parser = new Asn1BerParser(stream);
    Asn1BerNode* node;
    
    node = parser->parse();
    
    QVERIFY(node->tagClass() == CLASS_APPLICATION);
    QVERIFY(node->isClassApplication());
    QVERIFY(node->tag() == 12);
    QVERIFY(node->len() == 2);
    QVERIFY(node->val() == QByteArray("ab"));
}

void Asn1BerParserTest::testFile()
{
    QFile* file = new QFile("test_1.ber");
    file->open(QIODevice::ReadOnly);
        
    Asn1BerParser* parser = new Asn1BerParser(file);
    Asn1BerNode* node;
    
    node = parser->parse();
    
    
    QVERIFY(node->len() == 128);
    
    node = node->firstChild();
    
    QVERIFY(node->isClassContext());
    QVERIFY(!(node->isClassApplication()));
    QVERIFY(!(node->isClassUniversal()));
    QVERIFY(!(node->isClassPrivate()));
    QVERIFY(node->tag() == 1048608);
    QVERIFY(node->parent()->next()->next() == 0);
}

void Asn1BerParserTest::testFileToArray()
{
    QFile* file = new QFile("test_1.ber");
    file->open(QIODevice::ReadOnly);
        
    Asn1BerParser* parser = new Asn1BerParser(file->readAll());
    Asn1BerNode* node;
    
    node = parser->parse();
    
    
    QVERIFY(node->len() == 128);
    
    node = node->firstChild();
    
    QVERIFY(node->isClassContext());
    QVERIFY(!(node->isClassApplication()));
    QVERIFY(!(node->isClassUniversal()));
    QVERIFY(!(node->isClassPrivate()));
    QVERIFY(node->tag() == 1048608);
    QVERIFY(node->parent()->next()->next() == 0);
}

void Asn1BerParserTest::testComplexNoLength()
{
    QByteArray stream;
    
    stream.append((char) 0x30); // Universal [16] (sequence/sequence of)
    stream.append((char) UNKNOWN_LENGTH);
    stream.append((char) 0x30); // Universal [16] (sequence/sequence of)
    stream.append((char) UNKNOWN_LENGTH);
    stream.append((char) 0x40); // Application [0]
    stream.append((char) 0x01); // length = 1
    stream.append('a');
    stream.append((char) 0x41); // Application [1]
    stream.append((char) 0x01); // length = 1
    stream.append('b');
    stream.append((char) 0); // end-of-contents (2 octets)
    stream.append((char) 0);
    stream.append((char) 0x30); // Universal [16] (sequence/sequence of)
    stream.append((char) UNKNOWN_LENGTH);
    stream.append((char) 0x42); // Application [2]
    stream.append((char) 0x03); // length = 3
    stream.append('a');
    stream.append('c');
    stream.append('r');
    stream.append((char) 0x43); // Application [3]
    stream.append((char) 0x01); // length = 1
    stream.append('y');
    stream.append((char) 0); // end-of-contents (2 octets)
    stream.append((char) 0);
    stream.append((char) 0); // end-of-contents (2 octets)
    stream.append((char) 0);
    
    Asn1BerParser* parser = new Asn1BerParser(stream);
    Asn1BerNode* node;
    
    node = parser->parse();
    
    QVERIFY(node->tag() == 16);
    QVERIFY(!(node->hasDefinedLength()));
    QVERIFY(node->firstChild()->tag() == 16);
    QVERIFY(node->firstChild()->firstChild()->val() == QByteArray("a"));
    QVERIFY(node->firstChild()->next()->firstChild()->val() == QByteArray("acr"));
    QVERIFY(node->firstChild()->next()->firstChild()->next()->next() == 0);
    QVERIFY(node->next() == 0);

}

void Asn1BerParserTest::testZeroLength()
{
    QByteArray stream;
    stream.append((char) 0x40); // application 0 primitive
    stream.append((char) 0); // len = 0
    stream.append((char) 0x60); //application 0 constructed
    stream.append((char) 0); // len = 0
    stream.append((char) 0x41); //application 1 primitive
    stream.append((char) 2); // len = 2
    stream.append('a');
    stream.append('a');
    
    Asn1BerParser* parser = new Asn1BerParser(stream);
    Asn1BerNode* node;
    
    node = parser->parse();
    
    QVERIFY(node->tag() == 0);
    QVERIFY(node->len() == 0);
    QVERIFY(node->val().isEmpty());
    QVERIFY(node->isCodingPrimitive());
    QVERIFY(node->hasDefinedLength());
    
    node = node->next();
    
    QVERIFY(node->tag() == 0);
    QVERIFY(node->len() == 0);
    QVERIFY(node->isCodingComplex());
    QVERIFY(node->hasDefinedLength());
    QVERIFY(node->firstChild() == 0);
    
    node = node->next();
    
    QVERIFY(node->val() == QString("aa"));
}

void Asn1BerParserTest::cleanupTestCase()
{
    qDebug() << "cleanup";
}

// this thing defines a standard main function
QTEST_MAIN(Asn1BerParserTest)

#include "parsertest.moc"